module KDTree where

import Point

data KDtree a =
	KdNull
	|KdLeaf [Point a]
	|KdNode Bool a (KDtree a) (KDtree a)
	deriving Show

mais_prox ponto lista = [c | (a,b,c) <- [list] ]
	where list = minimum [ (cDist ponto b,ponto,b) | b <- lista]
--------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
-- Ordena uma lista de tuplas por x ou por y -----------------------------------------------------
-- qSortX - por x --------------------------------------------------------------------------------
-- qSortY - por y --------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
sSplitX op a [] = []
sSplitX op a x
	| op (xCord a) (xCord (head x)) = [(head x)] ++ (sSplitX op a (tail x))
	| otherwise = sSplitX op a (tail x)

qSortX [] = []
qSortX (x:xs) = qSortX (sSplitX (>) x xs) ++
		[x] ++
		qSortX (sSplitX (<=) x xs)
--------------------------------------------------------------------------------------------------
sSplitY op a [] = []
sSplitY op a x
	| op (yCord a) (yCord (head x)) = [(head x)] ++ (sSplitY op a (tail x))
	| otherwise = sSplitY op a (tail x)

qSortY [] = []
qSortY (x:xs) = qSortY (sSplitY (>) x xs) ++
		[x] ++
		qSortY (sSplitY (<=) x xs)
--------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
-----Ex.1-----------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
changeBool::(Ord a)=>[Point a]->Bool->KDtree a
changeBool [] bool = KdNull
changeBool xs bool
        |(length xs) <= 8 = KdLeaf xs
        |otherwise = if bool
		then
			KdNode bool vGetDimT vTreeOneT vTreeTwoT
		else
			KdNode bool vGetDimF vTreeOneF vTreeTwoF
		where
			vTreeOneT = changeBool (vNovosPOT) (not bool)
			vTreeTwoT = changeBool (vNovosPTT) (not bool)
			vGetDimT  = xCord (last (take (div (length xs) 2) vListOrdX))
			vNovosPOT = [ pontos | pontos <- vListOrdX, (xCord pontos) <= vGetDimT]
			vNovosPTT = [ pontos | pontos <- vListOrdX, (xCord pontos) >  vGetDimT]
			vListOrdX = (qSortX xs)
			vTreeOneF = changeBool (vNovosPOF) (not bool)
			vTreeTwoF = changeBool (vNovosPTF) (not bool)
			vGetDimF  = yCord (last (take (div (length xs) 2) vListOrdY))
			vNovosPOF = [ pontos | pontos <- vListOrdY, (yCord pontos) <= vGetDimF]
			vNovosPTF = [ pontos | pontos <- vListOrdY, (yCord pontos) >  vGetDimF]
			vListOrdY = (qSortY xs)

build ::(Ord a)=>[Point a]->KDtree a
build [] = KdNull
build xs = changeBool xs True
--------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
-----Ex.2-----------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
-- 1 maior
-- 2 menor ou igual
--------------------------------------------------------------------------------------------------
isKDtree KdNull = True
isKDtree (KdLeaf xs) = True
isKDtree (KdNode bool tam ys xs) = aux_is_KDtree xs [(bool,tam,1)] && aux_is_KDtree ys [(bool,tam,2)]

aux_is_KDtree KdNull lista = True
aux_is_KDtree (KdLeaf xs) lista_cond = verify_inte xs lista_cond
aux_is_KDtree (KdNode bool tam ys xs) lista  = aux_is_KDtree xs ([(bool,tam,1)] ++ lista) &&
					       aux_is_KDtree ys ([(bool,tam,2)] ++ lista)

verify_inte xs lista_cond = and [ condicoes lista_cond ponto | ponto <- xs]

condicoes [ ] ponto = True
condicoes [b] ponto
		| prim b       = if ((terc b) == 1)
					then (fst ponto) > (segu b)
					else (fst ponto) <= (segu b)
		| not (prim b) = if ((terc b) == 1)
					then (snd ponto) > (segu b)
					else (snd ponto) <= (segu b)

condicoes ys  ponto
		| prim (head ys)       = if ((terc (head ys)) == 1)
						then (fst ponto) > (segu (head ys)) && condicoes (tail ys) ponto
						else (fst ponto) <= (segu (head ys)) && condicoes (tail ys) ponto
		| not (prim (head ys)) = if ((terc (head ys)) == 1)
						then (snd ponto) > (segu (head ys)) && condicoes (tail ys) ponto
						else (snd ponto) <= (segu (head ys)) && condicoes (tail ys) ponto
prim (a,b,c) = a
segu (a,b,c) = b
terc (a,b,c) = c
--------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
-----Ex.13-----------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------

--Deleta um elemento no vetor e devolve o vetor sem o elemento
deleteList::(Ord a)=>[Point a]->Point a->[Point a]
deleteList [] a = []
deleteList (x:xs) a
	| (x==a) = xs
	| otherwise = x:(deleteList xs a)

delete::(Ord a)=>Point a->KDtree a->KDtree a
delete a KdNull = KdNull
delete a (KdLeaf xs) = if (lista_sem_a == [])
	then (KdNull)
	else (KdLeaf lista_sem_a)
		where
		   lista_sem_a = deleteList xs a
delete a (KdNode bool dim left right) =
	if bool then
		if (xCord(a) <= dim ) then (KdNode bool dim (delete a left) right)
		else (KdNode bool dim left (delete a right))
	else
		if (yCord(a) <= dim ) then (KdNode bool dim (delete a left) right)
		else (KdNode bool dim left (delete a right))


findnn pt1 KdNull = pt1
findnn pt1 (KdLeaf xs) = head(mais_prox pt1 xs)
findnn pt1 (KdNode bool dim xs ys) =
	if bool
	then if (fst pt1 <= dim)
		then if (vDisP1P2 <= vDifCorX)
			then pt2
			else (me_dist pt1 pt2 pt3)
		else if (vDisP1P3 <= vDifCorX)
			then pt3
			else (me_dist pt1 pt3 pt2)
	else if (snd pt1 <= dim)
				then if (vDisP1P2 <= vDifCorY)
				then pt2
				else (me_dist pt1 pt2 pt3)
			else if (vDisP1P3 <= vDifCorY)
				then pt3
				else (me_dist pt1 pt3 pt2)
		where
			pt2 	 = findnn pt1 xs
			pt3 	 = findnn pt1 ys
			vDisP1P3 = cDist pt1 pt3
			vDisP1P2 = cDist pt1 pt2
			vDifCorX = abs(dim - fst pt1)
			vDifCorY = abs(dim - snd pt1)
me_dist pt1 pt2 pt3
	| (cDist pt1 pt2) <= (cDist pt1 pt3) = pt2
	| otherwise = pt3

